/*
 * main browser
 *
 * Copyright(c) 2008 olyutorskii
 * $Id: JdfBrowser.java 65 2008-07-03 09:07:24Z olyutorskii $
 */

package jp.sourceforge.jindolf;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.Action;
import javax.swing.JEditorPane;
import javax.swing.JPopupMenu;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.HyperlinkListener;
import javax.swing.plaf.TextUI;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Document;
import javax.swing.text.EditorKit;
import javax.swing.text.Highlighter;
import javax.swing.text.TextAction;
import javax.swing.text.View;

/**
 * 独自HTMLベースのメインブラウザ。
 * EditorKitはJdfEditorKitを利用する。
 */
@SuppressWarnings("serial")
public class JdfBrowser
        extends JEditorPane
        implements CaretListener{

    private static final Highlighter.HighlightPainter regexPainter =
            new DefaultHighlighter.DefaultHighlightPainter(Color.YELLOW);
    
    private final JPopupMenu popup = new JPopupMenu();
    private final TextAction copyAction;
    private final TextAction selAllAction;
    private String internalText = "";
    private Village village;
    private Period period;
    private TopicFilter topicFilter = null;
    private TopicFilter.FilterContext filterContext;
    private boolean hasShown = false;
    private Pattern searchRegex = null;

    /**
     * 初期Periodを指定してメインブラウザを作成する。
     * @param period 初期Period
     */
    public JdfBrowser(Period period){
        super();

        setEditable(false);
        setContentType("text/html");
        putClientProperty(JEditorPane.W3C_LENGTH_UNITS, Boolean.TRUE);

        copyAction = new JdfEditorKit.CopyAction();
        selAllAction = new TextAction(""){
            public void actionPerformed(ActionEvent event){
                selectAll();
            }
        };
        copyAction.putValue(Action.NAME, "選択範囲をコピー");
        selAllAction.putValue(Action.NAME, "全選択");
        
        popup.add(copyAction);
        popup.addSeparator();
        popup.add(selAllAction);
        setComponentPopupMenu(popup);

        addCaretListener(this);
      
        setPeriod(period);
        
        return;
    }
    
    /**
     * 現在のPeriodを返す。
     * @return 現在のPeriod
     */
    public Period getPeriod(){
        return this.period;
    }
    
    /**
     * Periodを更新する。
     * 古いPeriodの表示内容は消える。
     * 新しいPeriodの表示内容はまだ反映されない。
     * 既存のHyperlinkListenerは登録を外される。
     * @param period 新しいPeriod
     */
    synchronized public void setPeriod(Period period){
        if(period == this.period) return;
        
        this.period = period;
        this.hasShown = false;
        this.filterContext = null;
        setText("");

        if(this.period == null) return;
        
        if(period.getVillage() == this.village) return;
        this.village = period.getVillage();

        EditorKit kit = new JdfEditorKit(this.village);
        setEditorKit(kit);

        HyperlinkListener[] listeners = getHyperlinkListeners();
        for(HyperlinkListener listener : listeners){
            removeHyperlinkListener(listener);
        }
        addHyperlinkListener(new AnchorListener(this.village));

        return;
    }
    
    /**
     * フィルタを適用してPeriodの内容を出力する。
     * @param filter フィルタ
     */
    synchronized public void showTopics(boolean force){
        if(force || ! this.hasShown){
            CharSequence html = this.period.buildJdfHTML();
            setHTML(html);
            this.hasShown = true;
            this.filterContext = null;
        }
        
        highlightRegex();
        filtering();

        return;
    }
    
    /**
     * 指定したHTML文書で表示を更新する。
     * @param html HTML文書
     */
    private void setHTML(CharSequence html){
        Reader reader;
        if(html != null) reader = new StringReader(html.toString());
        else             reader = new StringReader("");
        
        EditorKit kit = getEditorKit();
        Document doc = kit.createDefaultDocument();

        try{
            kit.read(reader, doc, 0);
        }catch(IOException e){
            assert false;
        }catch(BadLocationException e){
            assert false;
        }

        setDocument(doc);
        int docLength = doc.getLength();
        try{
            this.internalText = doc.getText(0, docLength);
        }catch(BadLocationException e){
            assert false;
        }

        return;
    }

    public void setTopicFilter(TopicFilter filter){
        this.topicFilter = filter;
        return;
    }
    
    /**
     * 発言フィルタを適用する。
     * @param filter フィルタ
     */
    synchronized public void filtering(){
        if(   this.topicFilter != null
           && this.topicFilter.isSame(this.filterContext)){
            return;
        }

        List<TopicView> viewList = getTopicViewList(null, null);
        for(TopicView view : viewList){
            if(   this.topicFilter != null
               && this.topicFilter.isFiltered(view)){
                view.setVisible(false);
            }else{
                view.setVisible(true);
            }
            
            View parentView = view.getParent();
            if(parentView == null) continue;
            parentView.preferenceChanged(view, true, true);
        }

        if(this.topicFilter != null){
            this.filterContext = this.topicFilter.getFilterContext();
        }else{
            this.filterContext = null;
        }
        
        return;
    }
    
    /**
     * 検索パターンを設定する。
     * @param searchRegex 検索パターン
     */
    public void setSearchRegex(Pattern searchRegex){
        if(this.searchRegex == searchRegex) return;
        this.searchRegex = searchRegex;
        return;
    }
    
    /**
     * 与えられた正規表現にマッチする文字列をハイライト描画する。
     * @param pattern 正規表現
     */
    synchronized public void highlightRegex(){
        clearHighlight();
        if(this.searchRegex == null) return;

        Highlighter highlighter = getHighlighter();
        int docLength = this.internalText.length();
        Matcher matcher = this.searchRegex.matcher(this.internalText);
        matcher.region(0, docLength);
        while( matcher.find() ){
            int startPos = matcher.start();
            int endPos = matcher.end();
            if(startPos == endPos) break;
            matcher.region(endPos, docLength);
            try{
                highlighter.addHighlight(startPos, endPos, regexPainter);
            }catch(BadLocationException e){
                assert false;
                return;
            }
        }

        return;
    }
    
    /**
     * 検索ハイライトをクリアする。
     */
    private void clearHighlight(){
        Highlighter highlighter = getHighlighter();
        Highlighter.Highlight[] highlights = highlighter.getHighlights();
        for(Highlighter.Highlight highlight : highlights){
            if(highlight.getPainter() != regexPainter) continue;
            highlighter.removeHighlight(highlight);
        }
        return;
    }
    
    /**
     * ルートViewを得る。
     * @return ルートView
     */
    private View getRootView(){
        TextUI textUI = getUI();
        View root = textUI.getRootView(this);
        return root;
    }
    
    /**
     * 全てのTopicViewを与えられた親から再帰的に抽出する。
     * @param parent 親View
     * @param list TopicViewのリスト
     * @return TopicViewのリスト
     */
    private List<TopicView> getTopicViewList(View parent,
                                              List<TopicView> list){
        if(parent == null){
            parent = getRootView();
            list = new LinkedList<TopicView>();
        }
        
        int childNo = parent.getViewCount();
        for(int ct=0; ct <= childNo-1; ct++){
            View child = parent.getView(ct);
            if(child instanceof TopicView){
                list.add( (TopicView) child );
            }else{
                getTopicViewList(child, list);
            }
        }
        
        return list;
    }

    /**
     * カレット変更イベント（範囲選択操作）の受信に応じて
     * ポップアップメニューの内訳を変更する。
     * @param event カレット変更イベント
     */
    public void caretUpdate(CaretEvent event){
        int dot  = event.getDot();
        int mark = event.getMark();
        if(dot == mark) copyAction.setEnabled(false);
        else            copyAction.setEnabled(true);
        return;
    }
    
    /**
     * 描画品質をカスタマイズするためのフック
     * @param g グラフィックスコンテキスト
     */
    // TODO フキダシの角丸だけでもアンチエイリアス指定しときたいなぁ
    @Override
    synchronized public void paint(Graphics g){
        Graphics2D g2 = (Graphics2D) g;
        g2.addRenderingHints(GUIUtils.getDefaultHints());
        super.paint(g);
        return;
    }
    
    /**
     * L&F を変更する。
     */
    @Override
    public void updateUI(){
        super.updateUI();

        highlightRegex();
        // TODO なぜかここでフィルタリング処理を行っても次回paint()に反映されない。
        
        return;
    }
}
